/*	inline.h

	This file contains all the basic functions that would normally
	be inlined.  Notice that inlining can be explicitly overridden
	by defining the symbol DO_NOT_INLINE; this feature is used to
	make a non-inlined version of the library in papers.c, so that
	ordinary versions of the routines can be accessed by linking.

	1995 by Hank Dietz
*/


#ifdef	P_VAPERS


#ifndef DO_NOT_INLINE
extern inline
#endif
unsigned int
inb(unsigned short port)
{
	p_portmesg mesg;

	mesg.op = P_INB;
	mesg.from = IPROC;
	mesg.port = (port - P_PORTBASE);
	if (write(p_req_pipe, &mesg, sizeof(mesg)) !=
	    sizeof(mesg)) {
		exit(1);
	}
	if (read(p_res_pipe, &(mesg.value), sizeof(mesg.value)) !=
	    sizeof(mesg.value)) {
		exit(1);
	}
	return(mesg.value);
}

unsigned int
inb2()
{
	p_portmesg mesg;

	mesg.op = P_WAITRDY;
	mesg.from = IPROC;
	if (write(p_req_pipe, &mesg, sizeof(mesg)) !=
	    sizeof(mesg)) {
		exit(1);
	}
	if (read(p_res_pipe, &(mesg.value), sizeof(mesg.value)) !=
	    sizeof(mesg.value)) {
		exit(1);
	}
	return(mesg.value);
}


#ifndef DO_NOT_INLINE
extern inline
#endif
void
outb(unsigned char value,
unsigned short port)
{
	p_portmesg mesg;

	mesg.op = P_OUTB;
	mesg.from = IPROC;
	mesg.value = value;
	mesg.port = (port - P_PORTBASE);
	if (write(p_req_pipe, &mesg, sizeof(mesg)) !=
	    sizeof(mesg)) {
		exit(1);
	}
}


#else	/* ifndef P_VAPERS */

/*	Descriptions of the inb and outb instructions....
	These could have come from /usr/src/linux/include/asm/io.h,
	but we want things to stand alone, so, here they are....
*/


#ifndef DO_NOT_INLINE
extern inline
#endif
unsigned int
inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}


#ifndef DO_NOT_INLINE
extern inline
#endif
unsigned int
inb2()
{
	unsigned short port = ((unsigned short) (P_PORTBASE + 1));
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}


#ifndef DO_NOT_INLINE
extern inline
#endif
void
outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}


#endif


#ifndef DO_NOT_INLINE
extern inline
#endif
void
p_enqueue(register barrier mask)
{
	if ((mask != ALL_MASK) && (mask != OUR_MASK)) {
		p_error("general barrier partitioning is not supported");
	}

	last_mask = mask;
}


#ifndef DO_NOT_INLINE
extern inline
#endif
void
_p_wait(void)
{
	register portdata t;

	/* Flip the strobe bits... */
	t = (last_out ^ P_S);
	P_OUT(t);
	last_out = t;

	/* Wait for P_RDY state to match P_S state...  if we poll more than
	   twice, turn on the Red LED until the wait has been satisfied.
	*/
	if (t & P_S0) {
		if (!(P_IN() & P_RDY)) {
			if (!(P_IN() & P_RDY)) {
				P_OUT(t ^ (P_LG | P_LR));
				while (!(P_IN2() & P_RDY));
				P_OUT(t);
			}
		}
	} else {
		if (P_IN() & P_RDY) {
			if (P_IN() & P_RDY) {
				P_OUT(t ^ (P_LG | P_LR));
				while (P_IN2() & P_RDY);
				P_OUT(t);
			}
		}
	}
}


#ifndef DO_NOT_INLINE
extern inline
#endif
void
p_wait(void)
{
	/* Simple barrier synchronization on current barrier.
	   Nothing to do unless we are not alone...
	*/
	if (last_mask & ~OUR_MASK) {
		_p_wait();
	}
}


#ifndef DO_NOT_INLINE
extern inline
#endif
portdata
_p_waitvec(register portdata b)
{
	register int count = 0;
	register portdata x;
	register portdata t;

	/* First, change the data */
	t = ((last_out & (P_S | P_LG)) | b);
	P_OUT(t);

	/* Now flip the strobe bits... */
	P_OUT(t ^= P_S);
	last_out = t;

	/* Wait for P_RDY state to match P_S state...  if we poll more than
	   twice, turn on the Red LED until the wait has been satisfied.
	*/
	if (t & P_S0) {
		if (!((x = P_IN()) & P_RDY)) {
			if (!((x = P_IN()) & P_RDY)) {
				P_OUT(t ^ (P_LG | P_LR));
				while (!((x = P_IN2()) & P_RDY));
			}
		}
	} else {
		if ((x = P_IN()) & P_RDY) {
			if ((x = P_IN()) & P_RDY) {
				P_OUT(t ^ (P_LG | P_LR));
				while ((x = P_IN2()) & P_RDY);
			}
		}
	}

	/* Now do another sync to ensure everyone has seen the data...
	   notice that this has the side effect of turning the red LED
	   off if the above code left it on...
	*/
	_p_wait();

	/* return data bit vector */
	return(x & P_IN_MASK);
}


#ifndef DO_NOT_INLINE
extern inline
#endif
barrier
p_waitvec(register int flag)
{
	/* Do a barrier wait sending flag and return the collected flag
	   vector.  Notice that the NAND gates invert the data, so we must
	   undo that inversion either as we send it or as we recieve it;
	   this code inverts as we send, because that's essentially a
	   freebie.  If we're not the only PE in our mask, we have work to
	   do...  but if we are alone, we are done.
	*/
	if (last_mask & ~OUR_MASK) {
		/* Must gather a bit from each PE...  translate
		   {I3,I2,I1,I0} into a standard mask
		*/
		return(IToMask(_p_waitvec(flag ?
					  (ALL_MASK ^ OUR_MASK) :
					  ALL_MASK)));
	}

	/* Only us... */
	return(flag ? OUR_MASK : 0);
}


#ifndef DO_NOT_INLINE
extern inline
#endif
int
p_any(register int flag)
{
	/* Do a barrier wait sending flag and return 1 if any flag
	   was nonzero, 0 otherwise.
	*/
	if (last_mask & ~OUR_MASK) {
		/* A variation on p_waitvec()... */
		return(IToMask(_p_waitvec(flag ? 0xe : 0xf)));
	}

	/* Only us... */
	return(flag != 0);
}


#ifndef DO_NOT_INLINE
extern inline
#endif
int
p_all(register int flag)
{
	/* Do a barrier wait sending flag and return 1 if all flags
	   were nonzero, 0 otherwise.
	*/
	if (last_mask & ~OUR_MASK) {
		/* A variation on p_waitvec()... */
		return(!IToMask(_p_waitvec(flag ? 0xf : 0xe)));
	}

	/* Only us... */
	return(flag != 0);
}
