/*	intern.h

	Internal header file for PAPERS support code.

	1995 by Hank Dietz
*/


#ifndef _INTERN_H
#define _INTERN_H


#include "papers.h"


/*	Port Base address for the PAPERS interface....
	normally, this is 0x378, but it is 0x3bc for IBM ValuePoint
*/
#define	P_PORTBASE 0x378


#ifdef	P_VAPERS


/*	Stuff for the pipe-based simulation...
*/
typedef struct {
	uint8	op;
	uint8	from;
	uint8	port;
	uint8	value;
} p_portmesg;

#define P_INB	0
#define	P_WAITRDY 1
#define P_OUTB	2

extern	int	p_req_pipe;	/* Request pipe */
extern	int	p_res_pipe;	/* Response pipe */


/*	Macro to get permission to access parallel port...
	but for the simulation, nothing to do.
*/ 
#define	P_PERM()	/* Nothing */


#define	inline		/* Simulator code should NEVER inline */


#else	/* ifndef P_VAPERS */


/*	Macro to get permission to access parallel port...
	This can use either:

		ioperm(P_PORTBASE, 3, 1)
		iopl(3)

	The ioperm() call is safer, but yields slower port operations
	due to I/O protection checks and is Linux-specific.  Because
	the mode port shouldn't be accessed by user code, only 2, not
	3, port registers need to be accessible.
*/ 
#define	P_PERM() \
	if (ioperm(P_PORTBASE, 2, 1)) { \
		fprintf(stderr, \
			"PAPERS PE%d: can't access parallel port\n", \
			IPROC); \
		exit(1); \
	}


#endif


/*	Stuff concerning the regular output port...
*/
#define	P_OUT(x) \
	outb(((unsigned char)(x)), ((unsigned short) P_PORTBASE))
#define	P_S1	0x80		/* Strobe RDY -> 1 */
#define	P_S0	0x40		/* Strobe RDY -> 0 */
#define	P_S	(P_S0 | P_S1)	/* Strobe Mask */
#define	P_LR	0x20		/* LED Red */
#define	P_LG	0x10		/* LED Green */
#define	P_LY	(P_LG | P_LR)	/* LED Yellow (Red + Green) */
#define	P_LB	0x00		/* LED Black */
#define	P_D3	0x08		/* Data Bit 3 Value */
#define	P_D2	0x04		/* Data Bit 2 Value */
#define	P_D1	0x02		/* Data Bit 1 Value */
#define	P_D0	0x01		/* Data Bit 0 Value */


/*	Stuff concerning the input port...
	note that P_I3 is inverted in software to compensate
	for the port inverting it....
*/
#define	P_IN() \
	inb((unsigned short) (P_PORTBASE + 1))
#define	P_IN2() \
	inb2()
#define	P_RDY	0x80		/* Ready */
#define	P_I3	0x40		/* PE3 P_D bit (inverted) */
#define	P_I2	0x20		/* PE2 P_D bit */
#define	P_I1	0x10		/* PE1 P_D bit */
#define	P_I0	0x08		/* PE0 P_D bit */
#define	P_IN_MASK 0x78		/* Mask for all the above bits */


/*	Notice that the mode port isn't accessed by user code.
*/


typedef	unsigned char portdata;		/* Data type for port data */
extern portdata last_out;		/* Last output value */
extern barrier last_mask;		/* Last barrier mask */


/*	The following two macros and tables map {I3,I2,I1,I0} into a
	barrier mask and a barrier mask into {I3,I2,I1,I0}.
*/
#define	IToMask(d) (((d) >> 3) & 0xf)
#define	MaskToI(d) ((d) << 3)


#define	ISPE(x) (!((x)&~(NPROC-1)))	/* is x a valid PE number? */
#define	ISNOTPE(x) ((x)&~(NPROC-1))	/* is x NOT a valid PE number? */


#define	p_nand(x) IToMask(_p_waitvec((x) & 0xf))
#define	p_internal_vote(abstain, vote) \
	(0xf ^ ((abstain) ? p_nand(0xf) : p_nand(0xf ^ (1 << (vote)))))


#ifdef DO_NOT_INLINE

/*	Ok, now include the inline versions of all the basic support
	functions for PAPERS....  Notice that these routines can be
	somewhat bulky, which is why they are used as inline routines
	only within the internal support library.

	Only kidding...  the simulator doesn't need inlining.
*/

#include "inline.h"

#endif


#endif
