/*	libdemo.c

	This program is a useless little demo of all the TTL_PAPERS
	library functions.  The output is simply a series of pages,
	each of which describes a group of logically related functions.
	The functions are not tested, but are executed and timed, and
	the result is displayed as ANSI-sequence colored bar graphs
	showing the variation in execution time for each function
	across the processors.

	(This is, incidentally, lousy code.  ;-)

	1995 by Hank Dietz
*/


#include <stdio.h>
#include "paperslib.h"


#define	TIMEOP(opname, op) \
{ \
	register int i = (count / 5); \
	register double mytime; \
 \
	p_wait(); \
	p_timerStart(); \
	do { \
		{ op; } \
		{ op; } \
		{ op; } \
		{ op; } \
		{ op; } \
	} while (--i > 0); \
	mytime = p_reduceMind(p_timerElapsed() / (0.000001 * count)); \
	if (IPROC == CPROC) printf("%28s ", opname); \
	showbar(mytime, 1000.0); \
}


main(register int argc,
register char **argv,
register char **arge)
{
	static uint64 buf[NPROC];
	register void *vptr = &(buf[0]);
	register int count;

	/* Parse command line... */
	if (argc != 2) {
		fprintf(stderr,
			"Usage: %s <timing_loop_count>\n");
		exit(1);
	}
	count = atoi(argv[1]);
	if (argc < 1) {
		fprintf(stderr,
			"%s: <timing_loop_count> should be at least 1\n");
		exit(2);
	}


	/* Connect to the papers unit... */
	p_init();


	/* Run the benchmarks...
	   notice that the benchmark is for timing only;
	   the functionality is not really tested.
	*/

	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Library Demo\033[0m\n");
		printf("This program is a useless little demo of the TTL_PAPERS library\n");
		printf("functions.  The output is simply a series of \"pages,\" each of which\n");
		printf("describes a group of logically related functions.  The functions are\n");
		printf("not tested, but are executed and timed, and the result is displayed as\n");
		printf("ANSI-sequence colored bar graphs showing the variation in execution\n");
		printf("time for each function across the processors.  Each blue horizontal\n");
		printf("bar represents the benchmarked execution time.  The field of each\n");
		printf("graph is shown in red.  The scale is linear, with full scale being\n");
		printf("1,000us, or 1ms -- about the same time taken for a minimal Ethernet\n");
		printf("communication using a UNIX socket interface.\n");
	}

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Barrier Synchronization\033[0m\n");
		printf("Although TTL_PAPERS only implements static barrier synchronization, it is\n");
		printf("possible to statically partition the machine such that only one partition is\n");
		printf("active at a time.  Thus, there are three basic operations associated with\n");
		printf("barrier synchronization:  enqueuing a new barrier group (represented by a\n");
		printf("bit mask), waiting for a barrier across the current group, and determining a\n");
		printf("new barrier group based on evaluation of a conditional expression (parallel\n");
		printf("if).  The following functions provide these operations.\n");
		putchar('\n');
	}
TIMEOP("p_enqueue(m)",	p_enqueue(ALL_MASK))
TIMEOP("p_wait()",	p_wait())
TIMEOP("p_waitvec(f)",	p_waitvec(0))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS MIMD/SIMD/VLIW Mode Mixing\033[0m\n");
		printf("Because TTL_PAPERS provides very low latency synchronization and aggregate\n");
		printf("communication, it is possible not only to use the system for MIMD/SPMD\n");
		printf("execution, but also for SIMD (Single Instruction stream, Multiple Data\n");
		printf("stream) and and VLIW (Very Long Instruction Word) processing.  In these\n");
		printf("modes, the code is distributed across the processors, and appropriate\n");
		printf("synchronous aggregate operations are used to ensure proper SIMD or VLIW\n");
		printf("sequencing.  The result is not instruction-level parallelism, but parallel\n");
		printf("execution with a grain size of perhaps a dozen floating point operations.\n");
		printf("The p_waitvec() call is used to collect conditionals for multiway branches\n");
		printf("that often result from VLIW code scheduling.\n");
		putchar('\n');
	}
TIMEOP("p_wait()",	p_wait())
TIMEOP("p_waitvec(f)",	p_waitvec(0))
TIMEOP("p_any(f)",	p_any(0))
TIMEOP("p_all(f)",	p_all(0))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Voting Operations\033[0m\n");
		printf("Voting operations collect \"votes\" from all processors, generally so\n");
		printf("that a static schedule can be implemented.  Each processor obtains a\n");
		printf("bit mask with a 1 corresponding to each processor that voted for it.\n");
		printf("The vote count operation simply returns how many voted for us.\n");
		putchar('\n');
	}
TIMEOP("p_vote(pe)",		p_vote(0))
TIMEOP("p_voteCount(pe)",	p_voteCount(0))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Match Operations\033[0m\n");
		printf("Matching operations are very much like voting operations, but are\n");
		printf("different in that the returned bit mask has a 1 in each position\n");
		printf("corresponding to each processor that tried to match the same value\n");
		printf("that this processor matched.\n");
		putchar('\n');
	}
TIMEOP("p_match1u(d)",	p_match1u(0))
TIMEOP("p_match8(d)",	p_match8(0))
TIMEOP("p_match8u(d)",	p_match8u(0))
TIMEOP("p_match16(d)",	p_match16(0))
TIMEOP("p_match16u(d)",	p_match16u(0))
TIMEOP("p_match32(d)",	p_match32(0))
TIMEOP("p_match32u(d)",	p_match32u(0))
TIMEOP("p_match64(d)",	p_match64(0))
TIMEOP("p_match64u(d)",	p_match64u(0))

TIMEOP("p_matchBits8u(d, 1)",	p_matchBits8u(0, 1))
TIMEOP("p_matchBits16u(d, 1)",	p_matchBits16u(0, 1))
TIMEOP("p_matchBits32u(d, 1)",	p_matchBits32u(0, 1))
TIMEOP("p_matchBits64u(d, 1)",	p_matchBits64u(0, 1))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Match Count Operations\033[0m\n");
		printf("Match counting operations simply return the number of processors that\n");
		printf("posted the same match value.\n");
		putchar('\n');
	}
TIMEOP("p_matchCount1u(d)",	p_matchCount1u(0))
TIMEOP("p_matchCount8(d)",	p_matchCount8(0))
TIMEOP("p_matchCount8u(d)",	p_matchCount8u(0))
TIMEOP("p_matchCount16(d)",	p_matchCount16(0))
TIMEOP("p_matchCount16u(d)",	p_matchCount16u(0))
TIMEOP("p_matchCount32(d)",	p_matchCount32(0))
TIMEOP("p_matchCount32u(d)",	p_matchCount32u(0))
TIMEOP("p_matchCount64(d)",	p_matchCount64(0))
TIMEOP("p_matchCount64u(d)",	p_matchCount64u(0))

TIMEOP("p_matchCountBits8u(d, 1)",	p_matchCountBits8u(0, 1))
TIMEOP("p_matchCountBits16u(d, 1)",	p_matchCountBits16u(0, 1))
TIMEOP("p_matchCountBits32u(d, 1)",	p_matchCountBits32u(0, 1))
TIMEOP("p_matchCountBits64u(d, 1)",	p_matchCountBits64u(0, 1))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Gather\033[0m\n");
		printf("Gather operations collect data from all PEs into a local array such that\n");
		printf("data from PE i is placed in p[i].  This PE's contribution to the gather is\n");
		printf("d.  The suffixes on the various function names reflect the different data\n");
		printf("types, e.g., 8u means 8-bit unsigned integer.  Bits operations transfer only\n");
		printf("the specified number of bits for each datum.\n");
		putchar('\n');
	}
TIMEOP("p_gather1u(p, d)",	p_gather1u(vptr, 0))
TIMEOP("p_gather8(p, d)",	p_gather8(vptr, 0))
TIMEOP("p_gather8u(p, d)",	p_gather8u(vptr, 0))
TIMEOP("p_gather16(p, d)",	p_gather16(vptr, 0))
TIMEOP("p_gather16u(p, d)",	p_gather16u(vptr, 0))
TIMEOP("p_gather32(p, d)",	p_gather32(vptr, 0))
TIMEOP("p_gather32u(p, d)",	p_gather32u(vptr, 0))
TIMEOP("p_gather64(p, d)",	p_gather64(vptr, 0))
TIMEOP("p_gather64u(p, d)",	p_gather64u(vptr, 0))
TIMEOP("p_gatherf(p, d)",	p_gatherf(vptr, 0))
TIMEOP("p_gatherd(p, d)",	p_gatherd(vptr, 0))

TIMEOP("p_gatherBits8u(p, d, 1)",	p_gatherBits8u(vptr, 0, 1))
TIMEOP("p_gatherBits16u(p, d, 1)",	p_gatherBits16u(vptr, 0, 1))
TIMEOP("p_gatherBits32u(p, d, 1)",	p_gatherBits32u(vptr, 0, 1))
TIMEOP("p_gatherBits64u(p, d, 1)",	p_gatherBits64u(vptr, 0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Putget\033[0m\n");
		printf("Putget operations implement aggregate communication in which each PE can\n");
		printf("obtain the datum injected by the PE of its choice.  This isn't the usual\n");
		printf("message passing mechanism, and it isn't shared memory, but it achieves\n");
		printf("better hardware utilization and makes better use of compile-time information.\n");
		putchar('\n');
	}
TIMEOP("p_putget1u(d, s)",	p_putget1u(0, 0))
TIMEOP("p_putget8(d, s)",	p_putget8(0, 0))
TIMEOP("p_putget8u(d, s)",	p_putget8u(0, 0))
TIMEOP("p_putget16(d, s)",	p_putget16(0, 0))
TIMEOP("p_putget16u(d, s)",	p_putget16u(0, 0))
TIMEOP("p_putget32(d, s)",	p_putget32(0, 0))
TIMEOP("p_putget32u(d, s)",	p_putget32u(0, 0))
TIMEOP("p_putget64(d, s)",	p_putget64(0, 0))
TIMEOP("p_putget64u(d, s)",	p_putget64u(0, 0))
TIMEOP("p_putgetf(d, s)",	p_putgetf(0, 0))
TIMEOP("p_putgetd(d, s)",	p_putgetd(0, 0))

TIMEOP("p_putgetBits8u(d, s, 1)",	p_putgetBits8u(0, 0, 1))
TIMEOP("p_putgetBits16u(d, s, 1)",	p_putgetBits16u(0, 0, 1))
TIMEOP("p_putgetBits32u(d, s, 1)",	p_putgetBits32u(0, 0, 1))
TIMEOP("p_putgetBits64u(d, s, 1)",	p_putgetBits64u(0, 0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceAdd\033[0m\n");
		printf("ReduceAdd operations sum data values from all processors.\n");
		putchar('\n');
	}
TIMEOP("p_reduceAdd1u(d)",	p_reduceAdd1u(0))
TIMEOP("p_reduceAdd8(d)",	p_reduceAdd8(0))
TIMEOP("p_reduceAdd8u(d)",	p_reduceAdd8u(0))
TIMEOP("p_reduceAdd16(d)",	p_reduceAdd16(0))
TIMEOP("p_reduceAdd16u(d)",	p_reduceAdd16u(0))
TIMEOP("p_reduceAdd32(d)",	p_reduceAdd32(0))
TIMEOP("p_reduceAdd32u(d)",	p_reduceAdd32u(0))
TIMEOP("p_reduceAdd64(d)",	p_reduceAdd64(0))
TIMEOP("p_reduceAdd64u(d)",	p_reduceAdd64u(0))
TIMEOP("p_reduceAddf(d)",	p_reduceAddf(0))
TIMEOP("p_reduceAddd(d)",	p_reduceAddd(0))

TIMEOP("p_reduceAddBits8u(d, 1)",	p_reduceAddBits8u(0, 1))
TIMEOP("p_reduceAddBits16u(d, 1)",	p_reduceAddBits16u(0, 1))
TIMEOP("p_reduceAddBits32u(d, 1)",	p_reduceAddBits32u(0, 1))
TIMEOP("p_reduceAddBits64u(d, 1)",	p_reduceAddBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceMul\033[0m\n");
		printf("ReduceMul operations multiply data values from all processors.\n");
		putchar('\n');
	}
TIMEOP("p_reduceMul1u(d)",	p_reduceMul1u(0))
TIMEOP("p_reduceMul8(d)",	p_reduceMul8(0))
TIMEOP("p_reduceMul8u(d)",	p_reduceMul8u(0))
TIMEOP("p_reduceMul16(d)",	p_reduceMul16(0))
TIMEOP("p_reduceMul16u(d)",	p_reduceMul16u(0))
TIMEOP("p_reduceMul32(d)",	p_reduceMul32(0))
TIMEOP("p_reduceMul32u(d)",	p_reduceMul32u(0))
TIMEOP("p_reduceMul64(d)",	p_reduceMul64(0))
TIMEOP("p_reduceMul64u(d)",	p_reduceMul64u(0))
TIMEOP("p_reduceMulf(d)",	p_reduceMulf(0))
TIMEOP("p_reduceMuld(d)",	p_reduceMuld(0))

TIMEOP("p_reduceMulBits8u(d, 1)",	p_reduceMulBits8u(0, 1))
TIMEOP("p_reduceMulBits16u(d, 1)",	p_reduceMulBits16u(0, 1))
TIMEOP("p_reduceMulBits32u(d, 1)",	p_reduceMulBits32u(0, 1))
TIMEOP("p_reduceMulBits64u(d, 1)",	p_reduceMulBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceAnd\033[0m\n");
		printf("ReduceAnd operations bitwise AND data values from all processors.  This is\n");
		printf("particularly efficient in TTL_PAPERS because the hardware directly\n");
		printf("implements a 4-bit NAND of data across the processors.\n");
		putchar('\n');
	}
TIMEOP("p_reduceAnd1u(d)",	p_reduceAnd1u(0))
TIMEOP("p_reduceAnd8(d)",	p_reduceAnd8(0))
TIMEOP("p_reduceAnd8u(d)",	p_reduceAnd8u(0))
TIMEOP("p_reduceAnd16(d)",	p_reduceAnd16(0))
TIMEOP("p_reduceAnd16u(d)",	p_reduceAnd16u(0))
TIMEOP("p_reduceAnd32(d)",	p_reduceAnd32(0))
TIMEOP("p_reduceAnd32u(d)",	p_reduceAnd32u(0))
TIMEOP("p_reduceAnd64(d)",	p_reduceAnd64(0))
TIMEOP("p_reduceAnd64u(d)",	p_reduceAnd64u(0))

TIMEOP("p_reduceAndBits8u(d, 1)",	p_reduceAndBits8u(0, 1))
TIMEOP("p_reduceAndBits16u(d, 1)",	p_reduceAndBits16u(0, 1))
TIMEOP("p_reduceAndBits32u(d, 1)",	p_reduceAndBits32u(0, 1))
TIMEOP("p_reduceAndBits64u(d, 1)",	p_reduceAndBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceOr\033[0m\n");
		printf("ReduceOr operations bitwise OR data values from all processors.  This is\n");
		printf("particularly efficient in TTL_PAPERS because the hardware directly\n");
		printf("implements a 4-bit NAND of data across the processors.\n");
		putchar('\n');
	}
TIMEOP("p_reduceOr1u(d)",	p_reduceOr1u(0))
TIMEOP("p_reduceOr8(d)",	p_reduceOr8(0))
TIMEOP("p_reduceOr8u(d)",	p_reduceOr8u(0))
TIMEOP("p_reduceOr16(d)",	p_reduceOr16(0))
TIMEOP("p_reduceOr16u(d)",	p_reduceOr16u(0))
TIMEOP("p_reduceOr32(d)",	p_reduceOr32(0))
TIMEOP("p_reduceOr32u(d)",	p_reduceOr32u(0))
TIMEOP("p_reduceOr64(d)",	p_reduceOr64(0))
TIMEOP("p_reduceOr64u(d)",	p_reduceOr64u(0))

TIMEOP("p_reduceOrBits8u(d, 1)",	p_reduceOrBits8u(0, 1))
TIMEOP("p_reduceOrBits16u(d, 1)",	p_reduceOrBits16u(0, 1))
TIMEOP("p_reduceOrBits32u(d, 1)",	p_reduceOrBits32u(0, 1))
TIMEOP("p_reduceOrBits64u(d, 1)",	p_reduceOrBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceMin\033[0m\n");
		printf("ReduceMin operations return the minimum data value from any processor.\n");
		putchar('\n');
	}
TIMEOP("p_reduceMin1u(d)",	p_reduceMin1u(0))
TIMEOP("p_reduceMin8(d)",	p_reduceMin8(0))
TIMEOP("p_reduceMin8u(d)",	p_reduceMin8u(0))
TIMEOP("p_reduceMin16(d)",	p_reduceMin16(0))
TIMEOP("p_reduceMin16u(d)",	p_reduceMin16u(0))
TIMEOP("p_reduceMin32(d)",	p_reduceMin32(0))
TIMEOP("p_reduceMin32u(d)",	p_reduceMin32u(0))
TIMEOP("p_reduceMin64(d)",	p_reduceMin64(0))
TIMEOP("p_reduceMin64u(d)",	p_reduceMin64u(0))
TIMEOP("p_reduceMinf(d)",	p_reduceMinf(0))
TIMEOP("p_reduceMind(d)",	p_reduceMind(0))

TIMEOP("p_reduceMinBits8u(d, 1)",	p_reduceMinBits8u(0, 1))
TIMEOP("p_reduceMinBits16u(d, 1)",	p_reduceMinBits16u(0, 1))
TIMEOP("p_reduceMinBits32u(d, 1)",	p_reduceMinBits32u(0, 1))
TIMEOP("p_reduceMinBits64u(d, 1)",	p_reduceMinBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ReduceMax\033[0m\n");
		printf("ReduceMax operations return the maximum data value from any processor.\n");
		putchar('\n');
	}
TIMEOP("p_reduceMax1u(d)",	p_reduceMax1u(0))
TIMEOP("p_reduceMax8(d)",	p_reduceMax8(0))
TIMEOP("p_reduceMax8u(d)",	p_reduceMax8u(0))
TIMEOP("p_reduceMax16(d)",	p_reduceMax16(0))
TIMEOP("p_reduceMax16u(d)",	p_reduceMax16u(0))
TIMEOP("p_reduceMax32(d)",	p_reduceMax32(0))
TIMEOP("p_reduceMax32u(d)",	p_reduceMax32u(0))
TIMEOP("p_reduceMax64(d)",	p_reduceMax64(0))
TIMEOP("p_reduceMax64u(d)",	p_reduceMax64u(0))
TIMEOP("p_reduceMaxf(d)",	p_reduceMaxf(0))
TIMEOP("p_reduceMaxd(d)",	p_reduceMaxd(0))

TIMEOP("p_reduceMaxBits8u(d, 1)",	p_reduceMaxBits8u(0, 1))
TIMEOP("p_reduceMaxBits16u(d, 1)",	p_reduceMaxBits16u(0, 1))
TIMEOP("p_reduceMaxBits32u(d, 1)",	p_reduceMaxBits32u(0, 1))
TIMEOP("p_reduceMaxBits64u(d, 1)",	p_reduceMaxBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanAdd\033[0m\n");
		printf("ScanAdd is very similar to reduceAdd, but returns the sum of values from PEs\n");
		printf("0 to i in PE i, i.e., scanAdd performs parallel prefix add.\n");
		putchar('\n');
	}
TIMEOP("p_scanAdd1u(d)",	p_scanAdd1u(0))
TIMEOP("p_scanAdd8(d)",		p_scanAdd8(0))
TIMEOP("p_scanAdd8u(d)",	p_scanAdd8u(0))
TIMEOP("p_scanAdd16(d)",	p_scanAdd16(0))
TIMEOP("p_scanAdd16u(d)",	p_scanAdd16u(0))
TIMEOP("p_scanAdd32(d)",	p_scanAdd32(0))
TIMEOP("p_scanAdd32u(d)",	p_scanAdd32u(0))
TIMEOP("p_scanAdd64(d)",	p_scanAdd64(0))
TIMEOP("p_scanAdd64u(d)",	p_scanAdd64u(0))
TIMEOP("p_scanAddf(d)",		p_scanAddf(0))
TIMEOP("p_scanAddd(d)",		p_scanAddd(0))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanMul\033[0m\n");
		printf("ScanMul is very similar to reduceMul, but returns the product of values from\n");
		printf("PEs 0 to i in PE i, i.e., scanMul performs parallel prefix multiply.\n");
		putchar('\n');
	}
TIMEOP("p_scanMul1u(d)",	p_scanMul1u(0))
TIMEOP("p_scanMul8(d)",		p_scanMul8(0))
TIMEOP("p_scanMul8u(d)",	p_scanMul8u(0))
TIMEOP("p_scanMul16(d)",	p_scanMul16(0))
TIMEOP("p_scanMul16u(d)",	p_scanMul16u(0))
TIMEOP("p_scanMul32(d)",	p_scanMul32(0))
TIMEOP("p_scanMul32u(d)",	p_scanMul32u(0))
TIMEOP("p_scanMul64(d)",	p_scanMul64(0))
TIMEOP("p_scanMul64u(d)",	p_scanMul64u(0))
TIMEOP("p_scanMulf(d)",		p_scanMulf(0))
TIMEOP("p_scanMuld(d)",		p_scanMuld(0))

	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanAnd\033[0m\n");
		printf("ScanAnd performs parallel prefix bitwise AND.\n");
		putchar('\n');
	}
TIMEOP("p_scanAnd1u(d)",	p_scanAnd1u(0))
TIMEOP("p_scanAnd8(d)",		p_scanAnd8(0))
TIMEOP("p_scanAnd8u(d)",	p_scanAnd8u(0))
TIMEOP("p_scanAnd16(d)",	p_scanAnd16(0))
TIMEOP("p_scanAnd16u(d)",	p_scanAnd16u(0))
TIMEOP("p_scanAnd32(d)",	p_scanAnd32(0))
TIMEOP("p_scanAnd32u(d)",	p_scanAnd32u(0))
TIMEOP("p_scanAnd64(d)",	p_scanAnd64(0))
TIMEOP("p_scanAnd64u(d)",	p_scanAnd64u(0))

TIMEOP("p_scanAndBits8u(d, 1)",		p_scanAndBits8u(0, 1))
TIMEOP("p_scanAndBits16u(d, 1)",	p_scanAndBits16u(0, 1))
TIMEOP("p_scanAndBits32u(d, 1)",	p_scanAndBits32u(0, 1))
TIMEOP("p_scanAndBits64u(d, 1)",	p_scanAndBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanOr\033[0m\n");
		printf("ScanOr performs parallel prefix bitwise OR.\n");
		putchar('\n');
	}
TIMEOP("p_scanOr1u(d)",		p_scanOr1u(0))
TIMEOP("p_scanOr8(d)",		p_scanOr8(0))
TIMEOP("p_scanOr8u(d)",		p_scanOr8u(0))
TIMEOP("p_scanOr16(d)",		p_scanOr16(0))
TIMEOP("p_scanOr16u(d)",	p_scanOr16u(0))
TIMEOP("p_scanOr32(d)",		p_scanOr32(0))
TIMEOP("p_scanOr32u(d)",	p_scanOr32u(0))
TIMEOP("p_scanOr64(d)",		p_scanOr64(0))
TIMEOP("p_scanOr64u(d)",	p_scanOr64u(0))

TIMEOP("p_scanOrBits8u(d, 1)",	p_scanOrBits8u(0, 1))
TIMEOP("p_scanOrBits16u(d, 1)",	p_scanOrBits16u(0, 1))
TIMEOP("p_scanOrBits32u(d, 1)",	p_scanOrBits32u(0, 1))
TIMEOP("p_scanOrBits64u(d, 1)",	p_scanOrBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanMin\033[0m\n");
		printf("ScanMin performs parallel prefix minimum.\n");
		putchar('\n');
	}
TIMEOP("p_scanMin1u(d)",	p_scanMin1u(0))
TIMEOP("p_scanMin8(d)",		p_scanMin8(0))
TIMEOP("p_scanMin8u(d)",	p_scanMin8u(0))
TIMEOP("p_scanMin16(d)",	p_scanMin16(0))
TIMEOP("p_scanMin16u(d)",	p_scanMin16u(0))
TIMEOP("p_scanMin32(d)",	p_scanMin32(0))
TIMEOP("p_scanMin32u(d)",	p_scanMin32u(0))
TIMEOP("p_scanMin64(d)",	p_scanMin64(0))
TIMEOP("p_scanMin64u(d)",	p_scanMin64u(0))
TIMEOP("p_scanMinf(d)",		p_scanMinf(0))
TIMEOP("p_scanMind(d)",		p_scanMind(0))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS ScanMax\033[0m\n");
		printf("ScanMax performs parallel prefix maximum.\n");
		putchar('\n');
	}
TIMEOP("p_scanMax1u(d)",	p_scanMax1u(0))
TIMEOP("p_scanMax8(d)",		p_scanMax8(0))
TIMEOP("p_scanMax8u(d)",	p_scanMax8u(0))
TIMEOP("p_scanMax16(d)",	p_scanMax16(0))
TIMEOP("p_scanMax16u(d)",	p_scanMax16u(0))
TIMEOP("p_scanMax32(d)",	p_scanMax32(0))
TIMEOP("p_scanMax32u(d)",	p_scanMax32u(0))
TIMEOP("p_scanMax64(d)",	p_scanMax64(0))
TIMEOP("p_scanMax64u(d)",	p_scanMax64u(0))
TIMEOP("p_scanMaxf(d)",		p_scanMaxf(0))
TIMEOP("p_scanMaxd(d)",		p_scanMaxd(0))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Rank (sorting)\033[0m\n");
		printf("Rank operations essentially sort data values across the PEs, returning the\n");
		printf("position in the sorted list for each PE's datum.\n");
		putchar('\n');
	}
TIMEOP("p_rank1u(d)",	p_rank1u(0))
TIMEOP("p_rank8(d)",	p_rank8(0))
TIMEOP("p_rank8u(d)",	p_rank8u(0))
TIMEOP("p_rank16(d)",	p_rank16(0))
TIMEOP("p_rank16u(d)",	p_rank16u(0))
TIMEOP("p_rank32(d)",	p_rank32(0))
TIMEOP("p_rank32u(d)",	p_rank32u(0))
TIMEOP("p_rank64(d)",	p_rank64(0))
TIMEOP("p_rank64u(d)",	p_rank64u(0))
TIMEOP("p_rankf(d)",	p_rankf(0))
TIMEOP("p_rankd(d)",	p_rankd(0))

TIMEOP("p_rankBits8u(d, 1)",	p_rankBits8u(0, 1))
TIMEOP("p_rankBits16u(d, 1)",	p_rankBits16u(0, 1))
TIMEOP("p_rankBits32u(d, 1)",	p_rankBits32u(0, 1))
TIMEOP("p_rankBits64u(d, 1)",	p_rankBits64u(0, 1))


	keywait();
	if (IPROC == CPROC) {
		newpage();
		printf("\033[1mLinux, TTL_PAPERS Selection Functions\033[0m\n");
		printf("Last, and least, it is sometimes useful to sequentially number only the PEs\n");
		printf("in our barrier group, to determine which is the lowest-numbered PE in our\n");
		printf("group, or to simply pick any PE which is in our group.  Note that none of\n");
		printf("these functions actually uses TTL_PAPERS; they simply examine the current\n");
		printf("barrier mask.\n");
		putchar('\n');
	}
TIMEOP("p_population()",	p_population())
TIMEOP("p_enumerate()",		p_enumerate())
TIMEOP("p_selectFirst()",	p_selectFirst())
TIMEOP("p_selectOne()",		p_selectOne())

	newpage();

	/* Exit cleanly */
	p_exit(0);
	exit(0);
}


showbar(register double t,
register double tmax)
{
	static char buf[512];
	register char *p = &(buf[0]);
	register int i, j;

	if (IPROC != CPROC) return;

	sprintf(p,
" %g microseconds                                                  ",
		t);
	if (t > tmax) {
		t = tmax;
	}
	i = ((41 * (t / tmax)) + 0.5);
	printf("\033[1;37;44m"); /* Blue */
	for (j=0; j<42; ++j) {
		putchar(*(p++));
		if (j == i) printf("\033[1;37;41m"); /* Red */
	}
	if (i >= 42) {
		putchar('!');
	}
	printf("\033[0m\n");
}

newpage()
{
	putchar('\n');
}

keywait()
{
	p_wait();
	if (IPROC == CPROC) {
		putchar('\n');
		printf("Hit return to continue.\n");
		getchar();
	}
	p_wait();
}
