/*	musical.c

	Music calibration program for directly using PC speaker.

	The output of this program is a shell script that either
	automatically adjusts the value of P_TIMEBASE or simply
	reports that the current value is close enough....

	1995 by Hank Dietz
*/


#include "papers.h"
#include "psound.c"
#include "timer.h"


/*	Calibration using a 10 second tone...
	500 Hz on PE0
	1K  Hz on PE1
	2k  Hz on PE2
	4k  Hz on PE3
*/
long voices[] = {
	10000, 2000000, 1000000, 500000, 250000,
	0
};


main()
{
	register long *p = &(voices[0]);
	register double t;

	p_init();
	p_initSound();

	if (IPROC == CPROC) {
		printf("echo Starting P_TIMEBASE calibration test...\n");
	}
	p_wait();
	p_timerStart();
	while (*p) {
		register long w;

		p_wait();

		if (w = p[IPROC + 1]) {
			p_tone(*p, w);
		}
		p += 5;
	}
	p_wait();
	t = p_timerElapsed();

	if (IPROC == CPROC) {
		register int i = ((int) (P_TIMEBASE * (t / 10.0)));

		if ((i < (0.99 * P_TIMEBASE)) ||
		    (i > (1.01 * P_TIMEBASE))) {
			/* More than 1% error... fix it.... */
			printf("echo Changing P_TIMEBASE from %d to %d\n",
			       P_TIMEBASE,
			       i);
			printf("cp psound.c psound.c.old\n");
			printf("sed 's/^");
			printf("#define	P_TIMEBASE %d", P_TIMEBASE);
			printf("$/");
			printf("#define	P_TIMEBASE %d", i);
			printf("/' <psound.c.old >psound.c\n");
		} else {
			/* Less than 1% error...  ok as is.... */
			printf("echo P_TIMEBASE %d is close enough to %d\n",
			       P_TIMEBASE,
			       i);
		}
	}
	p_wait();

	p_exit(0);
}
