/*	papers.h

	Support for basic functions using PAPERS.

	This also defines NPROC, CPROC, and IPROC.  NPROC and IPROC
	function as they do in MPL, as the total number of PEs and the
	number of this PE.  CPROC is the number of the PE that is
	connected to the console -- in a cluster, the number of the PE
	where the parallel program was initiated.  In this release,
	CPROC is simply 0, but this may change in future releases.  In
	general, user programs should have only the CPROC PE access
	stdin, stdout, and stderr.  Thus, CPROC also effectively says
	which PE should be treated as the controlling PE.

	1995 by Hank Dietz
*/

#ifndef _PAPERS_H
#define _PAPERS_H


#include "types.h"	/* PAPERS data types */


#define	NPROC	4			/* Number of PEs */
#define	NXPROC	2			/* PEs in x dimension */
#define	NYPROC	2			/* PEs in y dimension */
#define	LNPROC	2			/* Bits needed to encode NPROC */
#define	LNXPROC	1			/* Bits needed to encode NXPROC */
#define	LNYPROC	1			/* Bits needed to encode NYPROC */
#define	ALL_MASK ((1 << NPROC)-1)	/* All PEs barrier mask */

#define CPROC	0			/* PAPERS Console PE number */

extern int iproc;			/* PAPERS PE number */
extern int ixproc;			/* X coordinate */
extern int iyproc;			/* Y coordinate */
extern barrier our_mask;		/* Our barrier mask */
#define IPROC	((const int) iproc)		/* Protected iproc */
#define	IXPROC	((const int) ixproc)		/* Protected ixproc */
#define	IYPROC	((const int) iyproc)		/* Protected iyproc */
#define	OUR_MASK ((const barrier) our_mask)	/* Protected our_mask */


extern void p_enqueue(register barrier mask);
extern void p_wait(void);
extern barrier p_waitvec(register int flag);
extern int p_any(int flag);
extern int p_all(int flag);

extern void p_init(void);
extern void p_error(register char *s);
extern void p_exit(register int ecode);

extern unsigned char (*p_signalHandler)(unsigned char);
extern unsigned char p_die(unsigned char x);
extern void p_signal(unsigned char (*fnptr)(unsigned char));


#endif
