/*	psound.c

	1995 by Hank Dietz
*/

#include <stdio.h>
#include "psound.h"


/*	WARNING:  You must empirically determine the value of
	P_TIMEBASE as the time, in nanoseconds, for the code
	in p_tone()'s loop to go through a typical iteration.
	Any errors in this value will affect the accuracy of
	both the note frequency and the duration....
*/
#define	P_TIMEBASE 3796


#ifdef	P_VAPERS


void
p_initSound(void)
{
	fprintf(stderr,
		"p_initSound(): can't access speaker port\n");
	exit(1);
}

void
p_tone(register long duration,
register long wavelength)
{
	/* No portable way to do this.... */
}


#else	/* ifndef P_VAPERS */


extern inline unsigned int
inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

extern inline void
outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}


void
p_initSound(void)
{
	if (ioperm(((unsigned short) 0x61), 1, 1)) {
		fprintf(stderr,
			"p_initSound(): can't access speaker port\n");
		exit(1);
	}
}

void
p_tone(register long duration,
register long wavelength)
{
	/* Play the specified tone using the PC speaker...  duration
	   is given as a number of milliseconds, wavelength is given
	   in nanoseconds.  A wavelength of 0 is taken to mean a quiet
	   wait...  which is accomplished by simply nullifying the
	   speaker bit toggling....
	*/
	static long wavepos = 0;
	register long durpos = 0;
	register unsigned short toggle = (wavelength ? 2 : 0);

	/* To avoid clicking sounds at note transitions,
	   carry over wavepos values from previous notes...
	   but limit to wavelength, because otherwise it
	   would try to catch up and would sound very bad
	*/
	if (wavepos > wavelength) {
		wavepos = wavelength;
	}

	while (duration) {
		/* Is it time to toggle the speaker? */
		if ((wavepos += P_TIMEBASE) >= wavelength) {
			/* Flip speaker bit... */
			outb((inb((unsigned short) 0x61) ^ toggle),
			     ((unsigned short) 0x61));
			wavepos -= wavelength;
		} else {
			/* Leave speaker bit unchanged... */
			outb((inb((unsigned short) 0x61)),
			     ((unsigned short) 0x61));
		}

		/* Have we counted another millisecond...? */
		if ((durpos += P_TIMEBASE) >= 1000000) {
			--duration;
			durpos -= 1000000;
		}
	}
}


#endif
