/*	putget.c

	PAPERS putget operations.

	Put a datum into PAPERS and get a datum from whichever PE we
	select.  If the source PE we read from is not a participant in
	the barrier, the result will be a zero value.
*/


#include "putget.h"
#include "intern.h"


#ifdef	P_VAPERS
static void
#else
extern inline void
#endif
p_put8u(register uint8 datum)
{
	p_nand(datum & 0xf);
	p_nand((datum >> 4) & 0xf);
}

#ifdef	P_VAPERS
static uint8
#else
extern inline uint8
#endif
p_get8u(void)
{
	register uint8 datum = (0xff ^ p_nand(0xf));
	return(datum ^ (p_nand(0xf) << 4));
}

#ifdef	P_VAPERS
static uint8
#else
extern inline void
#endif
p_void8u(void)
{
	/* Like doing a p_get8u() and ignoring the value */
	p_nand(0xf);
	p_nand(0xf);
}


/*	Basic object putget operations...
*/
#define	PUTGET(name, type, bitcnt) \
type \
name(register type datum, \
register int source) \
{ \
	/* If we're the only PE in our mask, we are done...  */ \
	if (last_mask == OUR_MASK) { \
		return(datum); \
	} else { \
		/* How much work to do?  Send only what's needed! */ \
		static type s, r; \
		register barrier who = 1; \
		register barrier senders = ALL_MASK;\
			/* p_vote((source == IPROC), source); */\
		register barrier sourcemask = (1 << (source));\
 \
		/* Set send & recieve data (in case I send to me) */ \
		r = (s = (datum)); \
 \
		/* Send each datum that somebody else wants */ \
		while (who <= senders) { \
			if (who & senders) { \
				if (OUR_MASK == who) { \
					/* Our turn to send data */ \
					register uint8 *p = ((uint8 *) &s); \
					register uint8 *e = (p + (bitcnt / 8)); \
					do { \
						p_put8u(*p); \
					} while (++p < e); \
				} else if (sourcemask == who) { \
					/* We just recieve data... */ \
					register uint8 *p = ((uint8 *) &r); \
					register uint8 *e = (p + (bitcnt / 8)); \
					do { \
						*p = p_get8u(); \
					} while (++p < e); \
				} else { \
					/* We pretend to recieve data... */ \
					register uint8 *p = ((uint8 *) &r); \
					register uint8 *e = (p + (bitcnt / 8)); \
					do { \
						p_void8u(); \
					} while (++p < e); \
				} \
			} \
 \
			who += who; \
		} \
 \
		return(r); \
	} \
}


PUTGET(p_putget8,    int8,  8)
PUTGET(p_putget8u,  uint8,  8)
PUTGET(p_putget16,   int16, 16)
PUTGET(p_putget16u, uint16, 16)
PUTGET(p_putget32,   int32, 32)
PUTGET(p_putget32u, uint32, 32)
PUTGET(p_putget64,   int64, 64)
PUTGET(p_putget64u, uint64, 64)
PUTGET(p_putgetf,      f32, 32)
PUTGET(p_putgetd,      f64, 64)


uint1
p_putget1u(register uint1 datum,
register int source)
{
	/* Assumes 32-bit ints... */
	p_putgetBits32u(datum, (source != 0), 1);
}


/*	Bit sequence putget operations...
*/
#define	PUTGETBITS(name, type) \
type \
name(register type datum, \
register int source, \
register int bits) \
{ \
	register typeunion(type) d; \
	datum &= ((1 << bits) - 1); \
	d.datum = datum; \
 \
	/* If we're not the only PE in our mask, work to do...  */ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = \
			((1 << (source & (NPROC-1))) & \
			 last_mask); \
		register typeof(d.bits) bit = 1; \
		register typeunion(type) t; \
		t.bits = 0; \
 \
		/* For each bit... */ \
		while (--bits >= 0) { \
			/* Get a 3-bit stripe of data */ \
			register barrier x = \
				p_waitvec(d.bits & bit); \
 \
			/* Convert bits... */ \
			if (x & mask) t.bits |= bit; \
 \
			/* Next bit... */ \
			bit <<= 1; \
		} \
 \
		return(t.datum); \
	} \
 \
	return(datum); \
}

PUTGETBITS(p_putgetBits8u,  uint8)
PUTGETBITS(p_putgetBits16u, uint16)
PUTGETBITS(p_putgetBits32u, uint32)
PUTGETBITS(p_putgetBits64u, uint64)
