/*	reduce.c

	PAPERS reduce operations.

	Reduce data from all enabled PEs.  Our contribution is datum.
*/


#include "gather.h"
#include "reduce.h"
#include "intern.h"


#define	REDUCEADD(name, type, bitcnt, suffix) \
type \
name(register type datum) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 0))); \
	p_gather##suffix(&(buf[0]), datum); \
	return(buf[0] + buf[1] + buf[2] + buf[3]); \
}

REDUCEADD(p_reduceAdd8,    int8,  8,  8)
REDUCEADD(p_reduceAdd8u,  uint8,  8,  8u)
REDUCEADD(p_reduceAdd16,   int16, 16, 16)
REDUCEADD(p_reduceAdd16u, uint16, 16, 16u)
REDUCEADD(p_reduceAdd32,   int32, 32, 32)
REDUCEADD(p_reduceAdd32u, uint32, 32, 32u)
REDUCEADD(p_reduceAdd64,   int64, 64, 64)
REDUCEADD(p_reduceAdd64u, uint64, 64, 64u)
REDUCEADD(p_reduceAddf,      f32, 32, f)
REDUCEADD(p_reduceAddd,      f64, 64, d)


#define	REDUCEADDBITS(name, type, suffix) \
type \
name(register type datum, \
register int bits) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 0))); \
	p_gatherBits##suffix(&(buf[0]), datum, bits); \
	return(buf[0] + buf[1] + buf[2] + buf[3]); \
}

REDUCEADDBITS(p_reduceAddBits8u,  uint8,  8u)
REDUCEADDBITS(p_reduceAddBits16u, uint16, 16u)
REDUCEADDBITS(p_reduceAddBits32u, uint32, 32u)
REDUCEADDBITS(p_reduceAddBits64u, uint64, 64u)

#define	REDUCEMUL(name, type, bitcnt, suffix) \
type \
name(register type datum) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 1))); \
	p_gather##suffix(&(buf[0]), datum); \
	return(buf[0] * buf[1] * buf[2] * buf[3]); \
}

REDUCEMUL(p_reduceMul8,    int8,  8,  8)
REDUCEMUL(p_reduceMul8u,  uint8,  8,  8u)
REDUCEMUL(p_reduceMul16,   int16, 16, 16)
REDUCEMUL(p_reduceMul16u, uint16, 16, 16u)
REDUCEMUL(p_reduceMul32,   int32, 32, 32)
REDUCEMUL(p_reduceMul32u, uint32, 32, 32u)
REDUCEMUL(p_reduceMul64,   int64, 64, 64)
REDUCEMUL(p_reduceMul64u, uint64, 64, 64u)
REDUCEMUL(p_reduceMulf,      f32, 32, f)
REDUCEMUL(p_reduceMuld,      f64, 64, d)

#define	REDUCEMULBITS(name, type, suffix) \
type \
name(register type datum, \
register int bits) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 1))); \
	p_gatherBits##suffix(&(buf[0]), datum, bits); \
	return(buf[0] * buf[1] * buf[2] * buf[3]); \
}

REDUCEMULBITS(p_reduceMulBits8u,  uint8,  8u)
REDUCEMULBITS(p_reduceMulBits16u, uint16, 16u)
REDUCEMULBITS(p_reduceMulBits32u, uint32, 32u)
REDUCEMULBITS(p_reduceMulBits64u, uint64, 64u)


#define	REDUCEAND(name, type, bitcnt) \
type \
name(register type datum) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register int bits = (bitcnt - 4); \
 \
		/* Four bits at a time, and = not nand */ \
		do { \
			datum &= ~(p_nand((datum >> bits) & 0xf) \
				   << bits); \
		} while ((bits -= 4) >= 0); \
	} \
 \
	return(datum); \
}

REDUCEAND(p_reduceAnd8,    int8,  8)
REDUCEAND(p_reduceAnd8u,  uint8,  8)
REDUCEAND(p_reduceAnd16,   int16, 16)
REDUCEAND(p_reduceAnd16u, uint16, 16)
REDUCEAND(p_reduceAnd32,   int32, 32)
REDUCEAND(p_reduceAnd32u, uint32, 32)
REDUCEAND(p_reduceAnd64,   int64, 64)
REDUCEAND(p_reduceAnd64u, uint64, 64)

#define	REDUCEANDBITS(name, type) \
type \
name(register type datum, \
register int bits) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = last_mask; \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* Not a 1 if somebody else is 0... */ \
				if (p_waitvec(0) & mask) { \
					/* Flip the bit... */ \
					datum ^= bit; \
				} \
			} else { \
				/* We've got a 0; tell others... */ \
				p_waitvec(1); \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

REDUCEANDBITS(p_reduceAndBits8u,  uint8)
REDUCEANDBITS(p_reduceAndBits16u, uint16)
REDUCEANDBITS(p_reduceAndBits32u, uint32)
REDUCEANDBITS(p_reduceAndBits64u, uint64)

#define	REDUCEOR(name, type, bitcnt) \
type \
name(register type datum) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register int bits = (bitcnt - 4); \
 \
		/* Four bits at a time, or = nand not */ \
		do { \
			datum |= (p_nand((~(datum >> bits)) & 0xf) \
				  << bits); \
		} while ((bits -= 4) >= 0); \
	} \
 \
	return(datum); \
}

REDUCEOR(p_reduceOr8,    int8,  8)
REDUCEOR(p_reduceOr8u,  uint8,  8)
REDUCEOR(p_reduceOr16,   int16, 16)
REDUCEOR(p_reduceOr16u, uint16, 16)
REDUCEOR(p_reduceOr32,   int32, 32)
REDUCEOR(p_reduceOr32u, uint32, 32)
REDUCEOR(p_reduceOr64,   int64, 64)
REDUCEOR(p_reduceOr64u, uint64, 64)

#define	REDUCEORBITS(name, type) \
type \
name(register type datum, \
register int bits) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = last_mask; \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* We've got a 1; do dummy barrier... */ \
				p_waitvec(1); \
			} else { \
				/* Only a 1 if somebody else is... */ \
				if (p_waitvec(0) & mask) { \
					datum |= bit; \
				} \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

REDUCEORBITS(p_reduceOrBits8u,  uint8)
REDUCEORBITS(p_reduceOrBits16u, uint16)
REDUCEORBITS(p_reduceOrBits32u, uint32)
REDUCEORBITS(p_reduceOrBits64u, uint64)


#define	REDUCEMIN(name, type, bitcnt, suffix) \
type \
name(register type min) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = min))); \
	p_gather##suffix(&(buf[0]), min); \
	if (buf[0] < min) min = buf[0]; \
	if (buf[1] < min) min = buf[1]; \
	if (buf[2] < min) min = buf[2]; \
	if (buf[3] < min) min = buf[3]; \
	return(min); \
}

REDUCEMIN(p_reduceMin8,    int8,  8,  8)
REDUCEMIN(p_reduceMin8u,  uint8,  8,  8u)
REDUCEMIN(p_reduceMin16,   int16, 16, 16)
REDUCEMIN(p_reduceMin16u, uint16, 16, 16u)
REDUCEMIN(p_reduceMin32,   int32, 32, 32)
REDUCEMIN(p_reduceMin32u, uint32, 32, 32u)
REDUCEMIN(p_reduceMin64,   int64, 64, 64)
REDUCEMIN(p_reduceMin64u, uint64, 64, 64u)
REDUCEMIN(p_reduceMinf,      f32, 32, f)
REDUCEMIN(p_reduceMind,      f64, 64, d)

#define	REDUCEMINBITS(name, type, suffix) \
type \
name(register type min, \
register int bits) \
{ \
	type buf[NPROC]; \
	min &= ((1 << (bits - 1)) - 1); \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = min))); \
	p_gatherBits##suffix(&(buf[0]), min, bits); \
	if (buf[0] < min) min = buf[0]; \
	if (buf[1] < min) min = buf[1]; \
	if (buf[2] < min) min = buf[2]; \
	if (buf[3] < min) min = buf[3]; \
	return(min); \
}

REDUCEMINBITS(p_reduceMinBits8u,  uint8,  8u)
REDUCEMINBITS(p_reduceMinBits16u, uint16, 16u)
REDUCEMINBITS(p_reduceMinBits32u, uint32, 32u)
REDUCEMINBITS(p_reduceMinBits64u, uint64, 64u)

#define	REDUCEMAX(name, type, bitcnt, suffix) \
type \
name(register type max) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = max))); \
	p_gather##suffix(&(buf[0]), max); \
	if (buf[0] > max) max = buf[0]; \
	if (buf[1] > max) max = buf[1]; \
	if (buf[2] > max) max = buf[2]; \
	if (buf[3] > max) max = buf[3]; \
	return(max); \
}

REDUCEMAX(p_reduceMax8,    int8,  8,  8)
REDUCEMAX(p_reduceMax8u,  uint8,  8,  8u)
REDUCEMAX(p_reduceMax16,   int16, 16, 16)
REDUCEMAX(p_reduceMax16u, uint16, 16, 16u)
REDUCEMAX(p_reduceMax32,   int32, 32, 32)
REDUCEMAX(p_reduceMax32u, uint32, 32, 32u)
REDUCEMAX(p_reduceMax64,   int64, 64, 64)
REDUCEMAX(p_reduceMax64u, uint64, 64, 64u)
REDUCEMAX(p_reduceMaxf,      f32, 32, f)
REDUCEMAX(p_reduceMaxd,      f64, 64, d)

#define	REDUCEMAXBITS(name, type, suffix) \
type \
name(register type max, \
register int bits) \
{ \
	type buf[NPROC]; \
	max &= ((1 << (bits - 1)) - 1); \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = max))); \
	p_gatherBits##suffix(&(buf[0]), max, bits); \
	if (buf[0] > max) max = buf[0]; \
	if (buf[1] > max) max = buf[1]; \
	if (buf[2] > max) max = buf[2]; \
	if (buf[3] > max) max = buf[3]; \
	return(max); \
}

REDUCEMAXBITS(p_reduceMaxBits8u,  uint8,  8u)
REDUCEMAXBITS(p_reduceMaxBits16u, uint16, 16u)
REDUCEMAXBITS(p_reduceMaxBits32u, uint32, 32u)
REDUCEMAXBITS(p_reduceMaxBits64u, uint64, 64u)
