/*	scan.c

	PAPERS scan operations.

	These functions perform the specified parallel prefix
	operations.  The current TTL_PAPERS implementation is
	somewhat sloppy in that it uses gather and then does
	everything locally...  but it works.

	1995 by Hank Dietz
*/


#include "gather.h"
#include "scan.h"
#include "intern.h"


#define	SCANADD(name, type, bitcnt, suffix) \
type \
name(register type datum) \
{ \
	type buf[NPROC]; \
	register type t; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 0))); \
	p_gather##suffix(&(buf[0]), datum); \
	t = buf[0]; \
	if (IPROC > 0) { \
		t += buf[1]; \
		if (IPROC > 1) { \
			t += buf[2]; \
			if (IPROC > 2) { \
				t += buf[3]; \
			} \
		} \
	} \
	return(t); \
}

SCANADD(p_scanAdd8,    int8,  8,  8)
SCANADD(p_scanAdd8u,  uint8,  8,  8u)
SCANADD(p_scanAdd16,   int16, 16, 16)
SCANADD(p_scanAdd16u, uint16, 16, 16u)
SCANADD(p_scanAdd32,   int32, 32, 32)
SCANADD(p_scanAdd32u, uint32, 32, 32u)
SCANADD(p_scanAdd64,   int64, 64, 64)
SCANADD(p_scanAdd64u, uint64, 64, 64u)
SCANADD(p_scanAddf,      f32, 32, f)
SCANADD(p_scanAddd,      f64, 64, d)

uint1
p_scanAdd1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanOrBits32u((datum != 0), 1));
}


#define	SCANMUL(name, type, bitcnt, suffix) \
type \
name(register type datum) \
{ \
	type buf[NPROC]; \
	register type t; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = 1))); \
	p_gather##suffix(&(buf[0]), datum); \
	t = buf[0]; \
	if (IPROC > 0) { \
		t *= buf[1]; \
		if (IPROC > 1) { \
			t *= buf[2]; \
			if (IPROC > 2) { \
				t *= buf[3]; \
			} \
		} \
	} \
	return(t); \
}

SCANMUL(p_scanMul8,    int8,  8,  8)
SCANMUL(p_scanMul8u,  uint8,  8,  8u)
SCANMUL(p_scanMul16,   int16, 16, 16)
SCANMUL(p_scanMul16u, uint16, 16, 16u)
SCANMUL(p_scanMul32,   int32, 32, 32)
SCANMUL(p_scanMul32u, uint32, 32, 32u)
SCANMUL(p_scanMul64,   int64, 64, 64)
SCANMUL(p_scanMul64u, uint64, 64, 64u)
SCANMUL(p_scanMulf,      f32, 32, f)
SCANMUL(p_scanMuld,      f64, 64, d)

uint1
p_scanMul1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanAndBits32u((datum != 0), 1));
}


#define	SCANAND(name, type, bitcnt) \
type \
name(register type datum) \
{ \
	register int bits = bitcnt; \
 \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = \
			(last_mask & (OUR_MASK - 1)); \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* Not a 1 if somebody else is 0... */ \
				if (p_waitvec(0) & mask) { \
					/* Flip the bit... */ \
					datum ^= bit; \
				} \
			} else { \
				/* We've got a 0; tell others... */ \
				p_waitvec(1); \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

SCANAND(p_scanAnd8,    int8,  8)
SCANAND(p_scanAnd8u,  uint8,  8)
SCANAND(p_scanAnd16,   int16, 16)
SCANAND(p_scanAnd16u, uint16, 16)
SCANAND(p_scanAnd32,   int32, 32)
SCANAND(p_scanAnd32u, uint32, 32)
SCANAND(p_scanAnd64,   int64, 64)
SCANAND(p_scanAnd64u, uint64, 64)

uint1
p_scanAnd1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanAndBits32u((datum != 0), 1));
}

#define	SCANANDBITS(name, type) \
type \
name(register type datum, \
register int bits) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = \
			(last_mask & (OUR_MASK - 1)); \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* Not a 1 if somebody else is 0... */ \
				if (p_waitvec(0) & mask) { \
					/* Flip the bit... */ \
					datum ^= bit; \
				} \
			} else { \
				/* We've got a 0; tell others... */ \
				p_waitvec(1); \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

SCANANDBITS(p_scanAndBits8u,  uint8)
SCANANDBITS(p_scanAndBits16u, uint16)
SCANANDBITS(p_scanAndBits32u, uint32)
SCANANDBITS(p_scanAndBits64u, uint64)

#define	SCANOR(name, type, bitcnt) \
type \
name(register type datum) \
{ \
	register int bits = bitcnt; \
 \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = \
			(last_mask & (OUR_MASK - 1)); \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* We've got a 1; do dummy barrier... */ \
				p_waitvec(1); \
			} else { \
				/* Only a 1 if somebody else is... */ \
				if (p_waitvec(0) & mask) { \
					datum |= bit; \
				} \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

SCANOR(p_scanOr8,    int8,  8)
SCANOR(p_scanOr8u,  uint8,  8)
SCANOR(p_scanOr16,   int16, 16)
SCANOR(p_scanOr16u, uint16, 16)
SCANOR(p_scanOr32,   int32, 32)
SCANOR(p_scanOr32u, uint32, 32)
SCANOR(p_scanOr64,   int64, 64)
SCANOR(p_scanOr64u, uint64, 64)

uint1
p_scanOr1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanOrBits32u((datum != 0), 1));
}

#define	SCANORBITS(name, type) \
type \
name(register type datum, \
register int bits) \
{ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = \
			(last_mask & (OUR_MASK - 1)); \
		register type bit = 1; \
 \
		/* One bit at a time from all PEs... */ \
		do { \
			if (datum & bit) { \
				/* We've got a 1; do dummy barrier... */ \
				p_waitvec(1); \
			} else { \
				/* Only a 1 if somebody else is... */ \
				if (p_waitvec(0) & mask) { \
					datum |= bit; \
				} \
			} \
 \
			/* Go to the next bit position... */ \
			bit += bit; \
		} while (--bits); \
	} \
 \
	return(datum); \
}

SCANORBITS(p_scanOrBits8u,  uint8)
SCANORBITS(p_scanOrBits16u, uint16)
SCANORBITS(p_scanOrBits32u, uint32)
SCANORBITS(p_scanOrBits64u, uint64)


#define	SCANMIN(name, type, bitcnt, suffix) \
type \
name(register type min) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = min))); \
	p_gather##suffix(&(buf[0]), min); \
	if (IPROC > 0) { \
		if (buf[0] < min) min = buf[0]; \
		if (IPROC > 1) { \
			if (buf[1] < min) min = buf[1]; \
			if (IPROC > 2) { \
				if (buf[2] < min) min = buf[2]; \
			} \
		} \
	} \
	return(min); \
}

SCANMIN(p_scanMin8,    int8,  8,  8)
SCANMIN(p_scanMin8u,  uint8,  8,  8u)
SCANMIN(p_scanMin16,   int16, 16, 16)
SCANMIN(p_scanMin16u, uint16, 16, 16u)
SCANMIN(p_scanMin32,   int32, 32, 32)
SCANMIN(p_scanMin32u, uint32, 32, 32u)
SCANMIN(p_scanMin64,   int64, 64, 64)
SCANMIN(p_scanMin64u, uint64, 64, 64u)
SCANMIN(p_scanMinf,      f32, 32, f)
SCANMIN(p_scanMind,      f64, 64, d)

uint1
p_scanMin1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanAndBits32u((datum != 0), 1));
}

#define	SCANMAX(name, type, bitcnt, suffix) \
type \
name(register type max) \
{ \
	type buf[NPROC]; \
	buf[0] = (buf[1] = (buf[2] = (buf[3] = max))); \
	p_gather##suffix(&(buf[0]), max); \
	if (IPROC > 0) { \
		if (buf[0] > max) max = buf[0]; \
		if (IPROC > 1) { \
			if (buf[1] > max) max = buf[1]; \
			if (IPROC > 2) { \
				if (buf[2] > max) max = buf[2]; \
			} \
		} \
	} \
	return(max); \
}

SCANMAX(p_scanMax8,    int8,  8,  8)
SCANMAX(p_scanMax8u,  uint8,  8,  8u)
SCANMAX(p_scanMax16,   int16, 16, 16)
SCANMAX(p_scanMax16u, uint16, 16, 16u)
SCANMAX(p_scanMax32,   int32, 32, 32)
SCANMAX(p_scanMax32u, uint32, 32, 32u)
SCANMAX(p_scanMax64,   int64, 64, 64)
SCANMAX(p_scanMax64u, uint64, 64, 64u)
SCANMAX(p_scanMaxf,      f32, 32, f)
SCANMAX(p_scanMaxd,      f64, 64, d)

uint1
p_scanMax1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_scanOrBits32u((datum != 0), 1));
}


#define	RANK(name, type, suffix) \
int \
name(register type datum) \
{ \
	type buf[NPROC]; \
	register int rank = 0; \
	register int i = 0; \
 \
	p_gather##suffix(&(buf[0]), datum); \
	do { \
		if (last_mask & (1 << i)) { \
			if ((buf[i] > datum) || \
			    ((i < IPROC) && (buf[i] == datum))) { \
				++rank; \
			} \
		} \
	} while (++i < NPROC); \
	return(rank); \
}

RANK(p_rank8,    int8,  8)
RANK(p_rank8u,  uint8,  8u)
RANK(p_rank16,   int16, 16)
RANK(p_rank16u, uint16, 16u)
RANK(p_rank32,   int32, 32)
RANK(p_rank32u, uint32, 32u)
RANK(p_rank64,   int64, 64)
RANK(p_rank64u, uint64, 64u)
RANK(p_rankf,      f32, f)
RANK(p_rankd,      f64, d)

int
p_rank1u(register uint1 datum)
{
	/* Assumes uint1 is 32-bit int */
	return(p_rankBits32u((datum != 0), 1));
}

#define	RANKBITS(name, type, suffix) \
int \
name(register type datum, \
register int bits) \
{ \
	type buf[NPROC]; \
	register int rank = 0; \
	register int i = 0; \
 \
	p_gatherBits##suffix(&(buf[0]), datum, bits); \
	do { \
		if (last_mask & (1 << i)) { \
			if ((buf[i] > datum) || \
			    ((i < IPROC) && (buf[i] == datum))) { \
				++rank; \
			} \
		} \
	} while (++i < NPROC); \
	return(rank); \
}

RANKBITS(p_rankBits8u,  uint8,  8u)
RANKBITS(p_rankBits16u, uint16, 16u)
RANKBITS(p_rankBits32u, uint32, 32u)
RANKBITS(p_rankBits64u, uint64, 64u)


static char popcount[16] = {
/*	0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4
};

int
p_population()
{
	/* How many PEs are in current barrier mask?
	   This is just a population count operation.
	*/
	return(popcount[last_mask]);
}


int
p_enumerate()
{
	/* Number the PEs in the current barrier mask...
	   this is also a population count operation.
	*/
	return(popcount[last_mask & (OUR_MASK - 1)]);
}


static char lowbitpos[16] = {
/*	0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	0, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0
};

int
p_selectFirst()
{
	/* Select any PE that is in the current barrier mask...
	   notice that this doesn't even require a sync.
	   It is also worth noting that, if nobody is in the
	   current barrier mask, this incorrectly returns 0.
	   Then again, I'm always in my barrier mask, so that
	   can never happen....
	*/
	return(lowbitpos[last_mask]);
}

int
p_selectOne()
{
	/* p_selectOne() is a clone of p_selectFirst()
	   because the cost is the same anyway.
	*/
	return(lowbitpos[last_mask]);
}
