/*	vote.c

	PAPERS voting and matching operations.

	These operations are intended primarily to be used for
	pseudo-static scheduling of runtime operations.

	1995 by Hank Dietz
*/


#include "vote.h"
#include "intern.h"


static char popcount[16] = {
/*	0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15 */
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4
};

barrier
p_vote(register int pe)
{
	register barrier votemask;

	if (iproc & 1) {
		votemask = p_waitvec(pe & 1);
	} else {
		votemask = ~p_waitvec(pe & 1);
	}

	if (iproc & 2) {
		votemask &= p_waitvec(pe & 2);
	} else {
		votemask &= ~p_waitvec(pe & 2);
	}

	return(votemask & p_waitvec(pe != P_ABSTAIN));
}


int
p_voteCount(register int pe)
{
	register barrier votemask;

	if (iproc & 1) {
		votemask = p_waitvec(pe & 1);
	} else {
		votemask = ~p_waitvec(pe & 1);
	}

	if (iproc & 2) {
		votemask &= p_waitvec(pe & 2);
	} else {
		votemask &= ~p_waitvec(pe & 2);
	}

	return(popcount[votemask & p_waitvec(pe != P_ABSTAIN)]);
}


#define	MATCH(name, type, bitcnt) \
barrier \
name(register type datum) \
{ \
	register int bits = bitcnt; \
	register barrier matchmask = last_mask; \
 \
	do { \
		if (datum & 1) { \
			matchmask &= p_waitvec(1); \
		} else { \
			matchmask &= ~p_waitvec(0); \
		} \
		datum >>= 1; \
	} while (--bits); \
 \
	return(matchmask); \
}

MATCH(p_match8,    int8,  8)
MATCH(p_match8u,  uint8,  8)
MATCH(p_match16,   int16, 16)
MATCH(p_match16u, uint16, 16)
MATCH(p_match32,   int32, 32)
MATCH(p_match32u, uint32, 32)
MATCH(p_match64,   int64, 64)
MATCH(p_match64u, uint64, 64)

barrier
p_match1u(register uint1 datum)
{
	return(datum ?
	       p_waitvec(1) :
	       (last_mask ^ p_waitvec(0)));
}


#define	MATCHCOUNT(name, type, bitcnt) \
int \
name(register type datum) \
{ \
	register int bits = bitcnt; \
	register barrier matchmask = last_mask; \
 \
	do { \
		if (datum & 1) { \
			matchmask &= p_waitvec(1); \
		} else { \
			matchmask &= ~p_waitvec(0); \
		} \
		datum >>= 1; \
	} while (--bits); \
 \
	return(popcount[matchmask]); \
}

MATCHCOUNT(p_matchCount8,    int8,  8)
MATCHCOUNT(p_matchCount8u,  uint8,  8)
MATCHCOUNT(p_matchCount16,   int16, 16)
MATCHCOUNT(p_matchCount16u, uint16, 16)
MATCHCOUNT(p_matchCount32,   int32, 32)
MATCHCOUNT(p_matchCount32u, uint32, 32)
MATCHCOUNT(p_matchCount64,   int64, 64)
MATCHCOUNT(p_matchCount64u, uint64, 64)

int
p_matchCount1u(register uint1 datum)
{
	return(popcount[datum ?
		p_waitvec(1) :
		(last_mask ^ p_waitvec(0))]);
}


#define	MATCHBITS(name, type) \
barrier \
name(register type datum, \
register int bits) \
{ \
	register barrier matchmask = last_mask; \
 \
	do { \
		if (datum & 1) { \
			matchmask &= p_waitvec(1); \
		} else { \
			matchmask &= ~p_waitvec(0); \
		} \
		datum >>= 1; \
	} while (--bits); \
 \
	return(matchmask); \
}

MATCHBITS(p_matchBits8u,  uint8)
MATCHBITS(p_matchBits16u, uint16)
MATCHBITS(p_matchBits32u, uint32)
MATCHBITS(p_matchBits64u, uint64)


#define	MATCHCOUNTBITS(name, type) \
int \
name(register type datum, \
register int bits) \
{ \
	register barrier matchmask = last_mask; \
 \
	do { \
		if (datum & 1) { \
			matchmask &= p_waitvec(1); \
		} else { \
			matchmask &= ~p_waitvec(0); \
		} \
		datum >>= 1; \
	} while (--bits); \
 \
	return(popcount[matchmask]); \
}

MATCHCOUNTBITS(p_matchCountBits8u,  uint8)
MATCHCOUNTBITS(p_matchCountBits16u, uint16)
MATCHCOUNTBITS(p_matchCountBits32u, uint32)
MATCHCOUNTBITS(p_matchCountBits64u, uint64)
